package gg.malloc.defense;

import org.bukkit.plugin.java.JavaPlugin;
import org.bukkit.plugin.RegisteredServiceProvider;
import org.bukkit.Location;
import org.bukkit.entity.Entity;
import org.bukkit.entity.EntityType;
import org.bukkit.World;
import org.bukkit.WorldCreator;

import java.util.logging.Logger;
import java.util.logging.Level;

import java.util.ArrayList;
import java.util.HashMap;

import gg.malloc.defense.model.Arena;
import gg.malloc.defense.model.Spawnpoint;
import gg.malloc.defense.model.Game;

import gg.malloc.defense.games.LinearGame;
import gg.malloc.defense.games.ScaledWaves;

import gg.malloc.defense.engine.GameRunner;

import gg.malloc.defense.commands.AddPlayerCommand;
import gg.malloc.defense.commands.SetStageCommand;
import gg.malloc.defense.commands.PlayerReadyCommand;

public class Plugin extends JavaPlugin {
  ArrayList<Arena> m_arenas = new ArrayList<>();
  ArrayList<Game> m_games = new ArrayList<>();
  HashMap<World, GameRunner> m_runningGames = new HashMap<>();

  private class TestSpawn implements Spawnpoint {
    Location m_location;

    public TestSpawn(Location location) {
      m_location = location;
    }

    @Override
    public Location getLocation() {
      return m_location;
    }

    @Override
    public String getName() {
      return "Mob Spawner";
    }

    @Override
    public String getID() {
      return "mob-spawner";
    }
  }

  public void debuginfo() {
    getLogger().info("Debug Info:");
  }

  @Override
  public void onEnable() {
    getLogger().info("Malloc Defense registered");
    getLogger().setLevel(Level.FINEST);
    setupDemoGame();
    getCommand("setstage").setExecutor(new SetStageCommand(this));
    getCommand("addplayer").setExecutor(new AddPlayerCommand(this));
    getCommand("debuginfo").setExecutor(new DebuginfoCommand(this));
    getCommand("ready").setExecutor(new PlayerReadyCommand(this));
  }

  public GameRunner getRunnerForWorld(World world) {
    GameRunner ret;
    if (m_runningGames.containsKey(world)) {
      ret = m_runningGames.get(world);
    } else {
      ret = new GameRunner(this, m_games.get(0), m_arenas.get(0));
      m_runningGames.put(world, ret);
      getServer().getPluginManager().registerEvents(new GameEventHandler(ret), this);
    }
    return ret;
  }

  void setupDemoGame() {
    getLogger().info("Setting up demo data'");
    World testWorld = getServer().getWorld("quarry");
    if (testWorld == null) {
      testWorld = new WorldCreator("quarry").generateStructures(false).createWorld();
    }
    Spawnpoint[] spawnpoints = new Spawnpoint[3];
    spawnpoints[0] = new TestSpawn(new Location(testWorld, -15, 80, -46));
    spawnpoints[1] = new TestSpawn(new Location(testWorld, -1, 80, -45));
    spawnpoints[2] = new TestSpawn(new Location(testWorld, 12, 81, -42));
    Spawnpoint bombTarget = new TestSpawn(new Location(testWorld, -20, 80, 31));
    m_arenas.add(new MemoryArena("Test Arena", testWorld, spawnpoints, bombTarget));
    m_games.add(new ScaledWaves());
  }
}
