package gg.malloc.defense.engine;

import gg.malloc.defense.model.Game;
import gg.malloc.defense.model.Arena;
import gg.malloc.defense.model.Spawner;
import gg.malloc.defense.model.Wave;

import gg.malloc.defense.Plugin;

import java.util.logging.Logger;

import org.bukkit.Bukkit;
import org.bukkit.Sound;
import org.bukkit.World;
import org.bukkit.boss.BarColor;
import org.bukkit.boss.BarStyle;
import org.bukkit.boss.BossBar;
import org.bukkit.event.player.PlayerTeleportEvent;
import org.bukkit.event.entity.EntityTargetEvent;
import org.bukkit.event.entity.EntityDamageEvent;
import org.bukkit.event.entity.EntityDamageByEntityEvent;
import org.bukkit.entity.Entity;
import org.bukkit.entity.LivingEntity;
import org.bukkit.entity.ArmorStand;
import org.bukkit.entity.EntityType;
import org.bukkit.entity.Player;
import org.bukkit.scheduler.BukkitTask;

public class GameRunner {
  Arena m_arena;
  Game m_game;
  Stage m_stage;
  Plugin m_plugin;

  BossBar m_gameBar = Bukkit.createBossBar("Malloc Defense", BarColor.PURPLE, BarStyle.SOLID);
  BossBar m_waveBar = Bukkit.createBossBar("Malloc Defense", BarColor.PURPLE, BarStyle.SOLID);
  BossBar m_bombBar = Bukkit.createBossBar("Bomb HP", BarColor.RED, BarStyle.SOLID);

  BukkitTask m_countdownTask;

  MobManager m_mobs;
  WaveManager m_waves;
  PlayerManager m_players;

  LivingEntity m_bombTarget = null;
  int m_bombHP = 100;

  Logger m_log;

  public enum Stage {
    Idle,
    Warmup,
    Countdown,
    Playing,
    GameOver
  }

  class WaveManager {
    int m_currentWaveNum = 0;
    int m_currentBatch = 0;
    Wave m_currentWave = null;
    Game m_game;

    WaveManager(Game game) {
      m_game = game;
    }

    void reset() {
      m_currentWaveNum = 0;
      m_currentBatch = 0;
      m_currentWave = null;
    }

    Wave currentWave() {
      return m_currentWave;
    }

    int currentWaveNum() {
      return m_currentWaveNum;
    }

    int currentBatchNum() {
      return m_currentBatch;
    }

    double progress() {
      return (double)m_currentWaveNum / (double)m_game.getWaveCount();
    }

    boolean isLastWave() {
      return m_currentWaveNum >= m_game.getWaveCount();
    }

    boolean isLastBatch() {
      return m_currentBatch >= m_currentWave.batchCount();
    }

    void nextBatch() {
      m_currentBatch += 1;
    }

    void next() {
      m_currentWaveNum += 1;
      m_currentBatch = 0;
      m_currentWave = m_game.getWave(m_currentWaveNum);
    }
  }


  public GameRunner(Plugin plugin, Game game, Arena arena) {
    m_plugin = plugin;
    m_game = game;
    m_arena = arena;
    m_stage = Stage.Idle;
    m_gameBar.setVisible(true);
    m_waveBar.setVisible(false);
    m_bombBar.setVisible(false);
    m_mobs = new MobManager(m_game);
    m_waves = new WaveManager(m_game);
    m_players = new PlayerManager();
    m_log = m_plugin.getLogger();
  }

  int m_warmupCountdown = 0;

  public void handleEntityRetargeting(EntityTargetEvent evt) {
    if (m_mobs.contains(evt.getEntity()) && evt.getReason() == EntityTargetEvent.TargetReason.FORGOT_TARGET) {
      evt.setTarget(m_bombTarget);
    }
  }

  public void handleEntityDamage(EntityDamageEvent evt) {
    if (evt.getEntity() == m_bombTarget && evt instanceof EntityDamageByEntityEvent) {
      EntityDamageByEntityEvent entityEvt = (EntityDamageByEntityEvent)evt;
      if (m_mobs.contains(entityEvt.getDamager())) {
        entityEvt.getDamager().setGlowing(true);
        m_bombHP -= 1;
        broadcastMessage("The bomb has been struck!");
        m_arena.getWorld().playSound(m_bombTarget.getLocation(), Sound.ENTITY_ZOMBIE_ATTACK_IRON_DOOR, 1.5f, 0.9f);
        updateMobBars();
        if (m_bombHP <= 0) {
          m_arena.getWorld().strikeLightningEffect(m_bombTarget.getLocation());
          requestTransition(Stage.GameOver);
        }
      }
      evt.setCancelled(true);
    } else {
      m_mobs.handleEntityDamage(evt);
    }
  }

  private void countdownTick() {
    if (m_warmupCountdown == 0) {
      requestTransition(Stage.Playing);
    } else {
      updateMobBars();
      broadcastMessage("Starting game in " + m_warmupCountdown);
      m_warmupCountdown--;
      m_countdownTask = m_plugin.getServer().getScheduler().runTaskLater(m_plugin, () -> {
        countdownTick();
      }, 20);
    }
  }

  private boolean enterIdle() {
    m_bombHP = 100;
    m_waves.reset();
    m_mobs.clear();
    m_players.requestTransitionForAll(PlayerManager.State.Idle);
    if (m_bombTarget != null) {
      m_bombTarget.remove();
      m_bombTarget = null;
    }
    if (m_countdownTask != null) {
      m_countdownTask.cancel();
      m_countdownTask = null;
    }
    return true;
  }

  private boolean enterWarmup() {
    m_waves.next();
    for(Player p : m_players.getPlayers()) {
        if (m_players.requestTransition(p, PlayerManager.State.Playing)) {
          p.teleport(m_arena.getWorld().getSpawnLocation(), PlayerTeleportEvent.TeleportCause.PLUGIN);
        }
        m_players.setReady(p, false);
    }
    broadcastTitle("Warmup", "Prepare yourself for wave " + m_waves.currentWaveNum());
    m_mobs.clear();
    return true;
  }

  public void togglePlayerReady(Player p) {
    setPlayerReady(p, !m_players.isReady(p));
  }

  public void setPlayerReady(Player p, boolean isReady) {
    m_players.setReady(p, isReady);
    if (m_players.isEveryoneReady()) {
      requestTransition(Stage.Countdown);
    }
  }

  private boolean enterCountdown() {
    m_warmupCountdown = 10;
    countdownTick();
    return true;
  }

  private boolean enterPlaying() {
    m_log.info("Starting wave " + m_waves.currentWaveNum());
    if (m_bombTarget == null) {
      m_bombTarget = (LivingEntity)m_arena.getWorld().spawnEntity(m_arena.bombTarget().getLocation(), EntityType.ARMOR_STAND);
    }
    ArmorStand bombStand = (ArmorStand)m_bombTarget;
    bombStand.setCustomName("Bomb Target");
    bombStand.setVisible(true);
    bombStand.setCustomNameVisible(true);
    bombStand.setGlowing(true);
    for(Player p : m_players.getPlayers()) {
        if (m_players.requestTransition(p, PlayerManager.State.Playing)) {
          p.teleport(m_arena.getWorld().getSpawnLocation(), PlayerTeleportEvent.TeleportCause.PLUGIN);
        }
    }
    spawnNextBatch();
    broadcastTitle("Wave " + m_waves.currentWaveNum());
    return true;
  }
  
  private boolean enterGameOver() {
    broadcastTitle("Game Over!");
    if (m_countdownTask != null) {
      m_countdownTask.cancel();
      m_countdownTask = null;
    }
    m_mobs.clear();
    for(Player p : m_players.getPlayers()) {
      removePlayer(p);
    }
    return true;
  }

  private void updateMobBars() {
    m_gameBar.setVisible(true);
    switch(m_stage) {
      case Idle:
        m_gameBar.setProgress(1.0);
        m_gameBar.setTitle("Waiting for players...");
        m_gameBar.setColor(BarColor.PURPLE);
        m_waveBar.setVisible(false);
        m_bombBar.setVisible(false);
        break;
      case Warmup:
        m_gameBar.setProgress(m_waves.progress());
        m_gameBar.setTitle("Wave " + m_waves.currentWaveNum() + " / " + m_game.getWaveCount());
        m_gameBar.setColor(BarColor.PURPLE);
        m_waveBar.setVisible(true);
        m_waveBar.setColor(BarColor.BLUE);
        m_waveBar.setTitle("Warmup - Waiting for players to get ready...");
        m_waveBar.setProgress(m_players.readyProgress());
        m_bombBar.setVisible(false);
        break;
      case Countdown:
        m_gameBar.setProgress(m_waves.progress());
        m_gameBar.setTitle("Wave " + m_waves.currentWaveNum() + " / " + m_game.getWaveCount());
        m_gameBar.setColor(BarColor.PURPLE);
        m_waveBar.setVisible(true);
        m_waveBar.setColor(BarColor.YELLOW);
        m_waveBar.setTitle("Wave starting!");
        m_waveBar.setProgress((double)m_warmupCountdown / (double)10);
        m_bombBar.setVisible(false);
        break;
      case Playing:
        m_gameBar.setProgress(m_waves.progress());
        m_gameBar.setTitle("Wave " + m_waves.currentWaveNum() + " / " + m_game.getWaveCount());
        m_gameBar.setColor(BarColor.PURPLE);
        if (m_mobs.createdMobs() > 0) {
          m_waveBar.setVisible(true);
          m_waveBar.setColor(BarColor.GREEN);
          m_waveBar.setTitle("Mobs remaining: " + m_mobs.remainingMobs());
          m_waveBar.setProgress(m_mobs.progress());
        } else {
          m_waveBar.setVisible(false);
        }
        m_bombBar.setVisible(true);
        m_bombBar.setProgress((double)m_bombHP / (double)100);
        break;
      case GameOver:
        m_gameBar.setColor(BarColor.RED);
        m_gameBar.setProgress(1.0);
        m_gameBar.setTitle("Game Over!");
        m_waveBar.setVisible(false);
        m_bombBar.setVisible(false);
        break;
    }
  }

  private void spawnNextBatch() {
    broadcastMessage("Spawning batch " + m_waves.currentBatchNum());
    Spawner spawner = new GameSpawner(m_arena, m_mobs, m_players, m_bombTarget);
    m_waves.currentWave().spawnBatch(spawner, m_waves.currentBatchNum());
    updateMobBars();
  }

  public void handlePlayerDeath(Player player) {
    if (m_players.requestTransition(player, PlayerManager.State.Dead)) {
      m_arena.getWorld().strikeLightningEffect(player.getLocation());
      if (!m_players.isAnyoneAlive()) {
        broadcastMessage("Everyone is dead :(");
        requestTransition(Stage.GameOver);
      } else {
        m_log.info("Remaining players " + m_players.remainingPlayers());
      }
    }
  }

  public void handleEntityDeath(Entity entity) {
    if (m_mobs.killMob(entity)) {
      broadcastMessage("Killed game entity " + entity);
      updateMobBars();
      if (m_mobs.remainingMobs() <= 3) {
        m_log.info("Starting next batch!");
        if (m_waves.isLastBatch()) {
          if (m_waves.isLastWave()) {
            requestTransition(Stage.GameOver);
          } else if (m_mobs.empty()) {
            requestTransition(Stage.Warmup);
          }
        } else {
          m_waves.nextBatch();
          spawnNextBatch();
        }
      } else {
        m_log.fine("Living mobs remaining: " + m_mobs.remainingMobs());
      }
    }
  }

  private boolean attemptTransition(Stage stage) {
    if (m_stage == stage) {
      return false;
    }
    m_log.info("Game state: " + stage);
    switch(stage) {
      case Idle:
        return enterIdle();
      case Warmup:
        return enterWarmup();
      case Countdown:
        return enterCountdown();
      case Playing:
        return enterPlaying();
      case GameOver:
        return enterGameOver();
    }
    return false;
  }
  
  private boolean validateTransition(Stage from, Stage to) {
    switch(from) {
      case Idle:
        return to == Stage.Warmup;
      case Warmup:
        return to == Stage.Playing || to == Stage.Idle || to == Stage.Countdown ;
      case Countdown:
        return to == Stage.Playing || to == Stage.Idle || to == Stage.Warmup;
      case Playing:
        return to == Stage.Warmup || to == Stage.GameOver || to == Stage.Idle;
      case GameOver:
        return to == Stage.Idle;
    }
    return false;
  }

  public boolean requestTransition(Stage stage) {
    if (!validateTransition(m_stage, stage)) {
      m_log.warning("Attemped illegal transition: " + m_stage + " -> " + stage);
      return false;
    }
    if (attemptTransition(stage)) {
      m_log.info("Game transition: " + m_stage + " -> " + stage);
      m_stage = stage;
      updateMobBars();
      return true;
    }
    m_log.severe("Failed to complete transition: " + m_stage + " -> " + stage);
    return false;
  }

  public void addPlayer(Player p) {
    m_players.addPlayer(p);
    m_gameBar.addPlayer(p);
    m_waveBar.addPlayer(p);
    m_bombBar.addPlayer(p);
    if (m_stage == Stage.Idle || m_stage == Stage.Warmup) {
      if (m_players.requestTransition(p, PlayerManager.State.Playing)) {
        p.teleport(m_arena.getWorld().getSpawnLocation(), PlayerTeleportEvent.TeleportCause.PLUGIN);
      }
      broadcastMessage(p.getName() + " has joined the game");
      if (m_stage == Stage.Idle) {
        requestTransition(Stage.Warmup);
      }
    }
  }

  public void removePlayer(Player p) {
    m_gameBar.removePlayer(p);
    m_waveBar.removePlayer(p);
    m_players.removePlayer(p);
    if (m_players.isEmpty()) {
      requestTransition(Stage.Idle);
    }
  }

  void broadcastTitle(String title) {
    broadcastTitle(title, "");
  }

  void broadcastTitle(String title, String subtitle) {
    for(Player p : m_players.getPlayers()) {
      p.sendTitle(title, subtitle, 10, 70, 20);
    }
  }

  void broadcastMessage(String string) {
    World world = m_arena.getWorld();
    for(Player p : world.getPlayers()) {
      p.sendMessage(string);
    }
  }

  void registerSpawnedMob(Entity entity) {
    m_mobs.addEntity(entity);
  }
}
